package com.x10hosting.nanohatakamachi.th_dnh.archiver;

import java.io.*;
import java.util.*;
import java.util.zip.*;

public class Compress {
	private static byte[] MAGIC_ARCHIVE = "PACK_FILE\u0000".getBytes(); //PACK_FILE
	private static byte[] MAGIC_COMPRESS = "COMPRESS_ZIP\u0000".getBytes(); //COMPRESS_ZIP
	private static BufferedReader STDIN = new BufferedReader(new InputStreamReader(System.in));
	
	private String dir,dest; private boolean override,verbose,compress;
	private ArrayList<File> files = new ArrayList<File>();
	private Compress(String dir, String dest, boolean override, boolean verbose,boolean compress){this.dir=dir;this.dest=dest;this.override=override;this.verbose=verbose;this.compress=compress;}
	private void compress(){
		try{
			if(!new File(dir).exists()) throw new GeneralError(dir+" does not exist");
			if(override&&new File(dest).exists()){
				System.out.print("Override "+dest+"? [y/n] (n) ");
				if(!STDIN.readLine().equals("y")) return;
			}
			File dir = new File(this.dir);
			if(!dir.isDirectory()) throw new GeneralError(dir.getPath()+" is not a directory.");
			
			if(compress) System.out.println("Compressing "+dir+"...");
			else System.out.println("Packing "+dir+"...");
			
			filelisting(dir);
			checklisting();
			
			int offset = calcheadersize();
			
			ArrayList<byte[]> compressed = null;
			int[] sizes = new int[files.size()];
			if(!compress) sizes = sizes();
			else {
				if(verbose) System.out.println("Compressing files...");
				compressed = new ArrayList<byte[]>();
				for(int i=0;i<files.size();i++){
					if(verbose) System.out.println(" - "+files.get(i).getName());
					ByteArrayOutputStream byte_out = new ByteArrayOutputStream();
					NDataOutputStream out = new NDataOutputStream(byte_out);
					byte_out.write(MAGIC_COMPRESS);
					out.writeInt((int)files.get(i).length());

					byte[] temp = new byte[(int)files.get(i).length()]; 
					new DataInputStream(new FileInputStream(files.get(i))).readFully(temp);
					
					ByteArrayOutputStream temp2 = new ByteArrayOutputStream(0); 
					DeflaterOutputStream temp3 = new DeflaterOutputStream(temp2, new Deflater(Deflater.BEST_COMPRESSION));
					temp3.write(temp);
					temp3.finish();
				    byte_out.write(temp2.toByteArray());
				    
				    
				    compressed.add(byte_out.toByteArray());
				    
				    sizes[i] = byte_out.size();
				}
			}
			
			FileOutputStream file_out = new FileOutputStream(dest);
			NDataOutputStream out = new NDataOutputStream(file_out);
			if(verbose) System.out.println("Writing header...");
			if(verbose) System.out.println(" - Writing magic number");
			out.write(MAGIC_ARCHIVE);
			if(verbose) System.out.println(" - Writing number of files");
			out.writeInt(files.size());
			if(verbose) System.out.println(" - Writing file list");
			
			for(int i=0;i<files.size();i++){
				if(verbose) System.out.println("  - "+files.get(i).getName());
				printstring(out,files.get(i).getName());
				out.writeInt(offset);
				out.writeInt(sizes[i]);
				offset+=sizes[i];
			}
			if(verbose) System.out.println("Writing files...");
			for(int i=0;i<files.size();i++){
				if(verbose) System.out.println(" - "+files.get(i).getName());
				if(!compress) {
					byte[] temp = new byte[(int)files.get(i).length()]; 
					new DataInputStream(new FileInputStream(files.get(i))).readFully(temp);
					file_out.write(temp);
				}
				else file_out.write(compressed.get(i));
			}
			//file_out.write("\nCompressed with th_dnh_archiver".getBytes()); //What am I doing? :V
			if(compress) System.out.println("Compressed.");
			else System.out.println("Packed.");
		} catch (GeneralError e) {
			if(e.getMessage()!=null)System.err.println(e.getMessage());
		} catch (FileNotFoundException e) {
			System.err.println("Unexpected Error: File not found.");
			e.printStackTrace();
		} catch (IOException e) {
			System.err.println("Unexpected Error: Low level IO error.");
			e.printStackTrace();
		}
	}
	
	private void filelisting(File d) throws GeneralError{
		if(!d.isDirectory()) throw new GeneralError("Internal error: "+d.getPath()+" is not a directory.");
		String[] temp = d.list();
		for(String x:temp){
			if(new File(d.getPath()+File.separatorChar+x).isDirectory()) {
				if(verbose) System.out.println("Browsing "+d.getPath()+File.separatorChar+x);
				filelisting(new File(d.getPath()+File.separatorChar+x));
			}
			else {
				if(verbose) System.out.println(d.getPath()+File.separatorChar+x+" found");
				files.add(new File(d.getPath()+File.separatorChar+x));
			}
		}
	}
	private void checklisting() throws GeneralError{
		File[] temp = new File[files.size()];
		files.toArray(temp);
		ArrayList<String> filenames = new ArrayList<String>();
		for(File x:temp){
			if(filenames.contains(x.getName())) throw new GeneralError("More than one file named "+x.getName()+" exists in "+dir);
			filenames.add(x.getName());
		}
	}
	private int calcheadersize(){
		File[] files = new File[this.files.size()];
		this.files.toArray(files);
		int temp = MAGIC_ARCHIVE.length+4+(files.length*12);
		for(File x:files){
			temp+=x.getName().getBytes().length+1;
		}
		return temp;
	}
	private void printstring(NDataOutputStream out, String str) throws IOException{
		str = str + "\u0000";
		out.writeInt(str.length());
		out.write(str.getBytes());
	}
	private int[] sizes(){
		File[] files = new File[this.files.size()];
		this.files.toArray(files);
		int[] temp = new int[files.length];
		for(int i = 0;i<files.length;i++){
			temp[i] = (int) files[i].length();
		}
		return temp;
	}
	
	public static void compress(String dir, String dest, boolean override, boolean verbose, boolean compress){
		new Compress(dir,dest,override,verbose,compress).compress();
	}
	
	@SuppressWarnings("serial")private static class GeneralError extends Exception{public GeneralError(String string) {super(string);}}
}
